// Copyright (C) 2022 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'dart:developer';

import 'package:flutter/material.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:form_bloc_example/app/app.dart';
import 'package:form_bloc_example/app/metadata.dart';
import 'package:form_bloc_example/constants.dart';
import 'package:wyatt_form_bloc/wyatt_form_bloc.dart';

class CategoryIndicator extends StatelessWidget {
  const CategoryIndicator(
      {Key? key, required this.field, required this.builder})
      : super(key: key);

  final String field;
  final Function(BuildContext context, FormDataState state) builder;

  Color computeColor(Metadata meta) {
    switch (meta.category) {
      case Category.perso:
        return Colors.blue;
      case Category.business:
        return Colors.red;
    }
  }

  @override
  Widget build(BuildContext context) {
    return BlocBuilder<FormDataCubit, FormDataState>(
      builder: ((context, state) {
        final meta = state.data.metadataOf(field).extra as Metadata;
        final color = computeColor(meta);

        return Row(
          children: [
            Container(
              height: 8,
              width: 8,
              decoration: BoxDecoration(
                color: color,
                borderRadius: BorderRadius.circular(8),
              ),
            ),
            const SizedBox(width: 20,),
            SizedBox(
              width: MediaQuery.of(context).size.width - 45,
              child: builder(context, state),
            )
          ],
        );
      }),
    );
  }
}

class _NameInput extends StatelessWidget {
  @override
  Widget build(BuildContext context) {
    return CategoryIndicator(
        field: formFieldName,
        builder: (context, state) {
          return TextField(
            onChanged: (name) => context
                .read<FormDataCubit>()
                .dataChanged(formFieldName, Name.dirty(name)),
            keyboardType: TextInputType.name,
            decoration: InputDecoration(
              labelText: 'name',
              helperText: '',
              errorText:
                  state.data.isNotValid(formFieldName) ? 'invalid name' : null,
            ),
          );
        });
  }
}

class _EmailInput extends StatelessWidget {
  @override
  Widget build(BuildContext context) {
    return CategoryIndicator(
        field: formFieldEmail,
      builder: (context, state) {
        return TextField(
          onChanged: (email) => context
              .read<FormDataCubit>()
              .dataChanged(formFieldEmail, Email.dirty(email)),
          keyboardType: TextInputType.emailAddress,
          decoration: InputDecoration(
            labelText: 'email',
            helperText: '',
            errorText:
                state.data.isNotValid(formFieldEmail) ? 'invalid email' : null,
          ),
        );
      },
    );
  }
}

class _PhoneInput extends StatelessWidget {
  @override
  Widget build(BuildContext context) {
    return CategoryIndicator(
        field: formFieldPhone,
      builder: (context, state) {
        return TextField(
          onChanged: (phone) => context
              .read<FormDataCubit>()
              .dataChanged(formFieldPhone, Phone.dirty(phone)),
          keyboardType: TextInputType.phone,
          decoration: InputDecoration(
            labelText: 'phone',
            helperText: '',
            errorText:
                state.data.isNotValid(formFieldPhone) ? 'invalid phone' : null,
          ),
        );
      },
    );
  }
}

class _SirenInput extends StatelessWidget {
  @override
  Widget build(BuildContext context) {
    return CategoryIndicator(
        field: formFieldName,
      builder: (context, state) {
        return TextField(
          onChanged: (siren) => context
              .read<FormDataCubit>()
              .dataChanged(formFieldSiren, Siren.dirty(siren)),
          keyboardType: TextInputType.number,
          decoration: InputDecoration(
            labelText: 'siren',
            helperText: '',
            errorText:
                state.data.isNotValid(formFieldSiren) ? 'invalid SIREN' : null,
          ),
        );
      },
    );
  }
}

class _IbanInput extends StatelessWidget {
  @override
  Widget build(BuildContext context) {
    return BlocBuilder<FormDataCubit, FormDataState>(
      builder: (context, state) {
        return TextField(
          onChanged: (iban) => context
              .read<FormDataCubit>()
              .dataChanged(formFieldIban, Iban.dirty(iban)),
          keyboardType: TextInputType.text,
          decoration: InputDecoration(
            labelText: 'iban',
            helperText: '',
            errorText:
                state.data.isNotValid(formFieldIban) ? 'invalid IBAN' : null,
          ),
        );
      },
    );
  }
}

class _CheckListInput extends StatelessWidget {
  @override
  Widget build(BuildContext context) {
    return BlocBuilder<FormDataCubit, FormDataState>(
      builder: (context, state) {
        final _input =
            state.data.validatorOf<ListOption<String>>(formFieldList);
        final _options = _input.value;

        return Column(
          mainAxisSize: MainAxisSize.min,
          children: [
            ListTile(
              title: const Text('Checkbox1'),
              trailing: Checkbox(
                  value: _options.contains('checkbox1'),
                  onChanged: (_) {
                    context.read<FormDataCubit>().dataChanged(
                          formFieldList,
                          _input.select('checkbox1'),
                        );
                  }),
            ),
            ListTile(
              title: const Text('Checkbox2'),
              trailing: Checkbox(
                  value: _options.contains('checkbox2'),
                  onChanged: (_) {
                    context.read<FormDataCubit>().dataChanged(
                          formFieldList,
                          _input.select('checkbox2'),
                        );
                  }),
            ),
            ListTile(
              title: const Text('Checkbox3 (default)'),
              trailing: Checkbox(
                  value: _options.contains('checkbox3'),
                  onChanged: (_) {
                    context.read<FormDataCubit>().dataChanged(
                          formFieldList,
                          _input.select('checkbox3'),
                        );
                  }),
            ),
          ],
        );
      },
    );
  }
}

class _RadioListInput extends StatelessWidget {
  @override
  Widget build(BuildContext context) {
    return BlocBuilder<FormDataCubit, FormDataState>(
      builder: (context, state) {
        final _input = state.data.validatorOf<TextString>(formFieldRadio);

        return Column(
          mainAxisSize: MainAxisSize.min,
          children: [
            ListTile(
              title: const Text('Radio1'),
              trailing: Radio<bool>(
                  groupValue: true,
                  value: _input.value == 'radio1',
                  onChanged: (_) {
                    context.read<FormDataCubit>().dataChanged(
                          formFieldRadio,
                          const TextString.dirty('radio1'),
                        );
                  }),
            ),
            ListTile(
              title: const Text('Radio2'),
              trailing: Radio<bool>(
                  groupValue: true,
                  value: _input.value == 'radio2',
                  onChanged: (_) {
                    context.read<FormDataCubit>().dataChanged(
                          formFieldRadio,
                          const TextString.dirty('radio2'),
                        );
                  }),
            ),
            ListTile(
              title: const Text('Radio3'),
              trailing: Radio<bool>(
                  groupValue: true,
                  value: _input.value == 'radio3',
                  onChanged: (_) {
                    context.read<FormDataCubit>().dataChanged(
                          formFieldRadio,
                          const TextString.dirty('radio3'),
                        );
                  }),
            ),
          ],
        );
      },
    );
  }
}

class _CheckHiddenInput extends StatelessWidget {
  @override
  Widget build(BuildContext context) {
    return ListTile(
      title: const Text('This input is not exported'),
      trailing: BlocBuilder<FormDataCubit, FormDataState>(
        builder: (context, state) {
          return Checkbox(
              value: state.data.validatorOf<Boolean>(formFieldHidden).value,
              onChanged: (v) {
                final value = v!; // state is false, so value can't be null

                context.read<FormDataCubit>().dataChanged(
                      formFieldHidden,
                      Boolean.dirty(value: value),
                    );
              });
        },
      ),
    );
  }
}

class _CheckIsProInput extends StatelessWidget {
  @override
  Widget build(BuildContext context) {
    return ListTile(
      title: const Text('Are you a pro?'),
      trailing: BlocBuilder<FormDataCubit, FormDataState>(
        builder: (context, state) {
          return Checkbox(
              value: state.data.validatorOf<Boolean>(formFieldPro).value,
              onChanged: (isPro) {
                final value = isPro!; // state is false, so value can't be null

                context.read<FormDataCubit>().dataChanged(
                      formFieldPro,
                      Boolean.dirty(value: value),
                    );

                if (value) {
                  context.read<FormDataCubit>().updateFormData(
                      App.getProFormData(),
                      operation: SetOperation.union);
                } else {
                  context.read<FormDataCubit>().updateFormData(
                      App.getNormalFormData(),
                      operation: SetOperation.intersection);
                }
              });
        },
      ),
    );
  }
}

class _SignUpButton extends StatelessWidget {
  @override
  Widget build(BuildContext context) {
    return BlocBuilder<FormDataCubit, FormDataState>(
      buildWhen: (previous, current) => previous.status != current.status,
      builder: (context, state) {
        return state.status.isSubmissionInProgress
            ? const CircularProgressIndicator()
            : ElevatedButton(
                onPressed: state.status.isValidated
                    ? () => context.read<FormDataCubit>().submitForm()
                    : null,
                child: const Text('SIGN UP'),
              );
      },
    );
  }
}

class _DebugButton extends StatelessWidget {
  @override
  Widget build(BuildContext context) {
    return BlocBuilder<FormDataCubit, FormDataState>(
      builder: (context, state) {
        return ElevatedButton(
          onPressed: () {
            log(state.toString());
          },
          child: const Text('DEBUG'),
        );
      },
    );
  }
}

class SignUpForm extends StatelessWidget {
  const SignUpForm({Key? key}) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return BlocListener<FormDataCubit, FormDataState>(
      listener: (context, state) {
        if (state.status.isSubmissionSuccess) {
          Navigator.of(context).pop(); // Never happens here
        } else if (state.status.isSubmissionFailure) {
          ScaffoldMessenger.of(context)
            ..hideCurrentSnackBar()
            ..showSnackBar(
              SnackBar(content: Text(state.errorMessage ?? 'Sign Up Failure')),
            );
        }
      },
      child: Align(
        alignment: const Alignment(0, -1 / 3),
        child: Column(
          mainAxisSize: MainAxisSize.min,
          children: [
            _NameInput(),
            const SizedBox(height: 8),
            _EmailInput(),
            const SizedBox(height: 8),
            _PhoneInput(),
            const SizedBox(height: 8),
            _CheckListInput(),
            const SizedBox(height: 8),
            _RadioListInput(),
            const SizedBox(height: 8),
            _CheckHiddenInput(),
            const SizedBox(height: 8),
            _CheckIsProInput(),
            const SizedBox(height: 8),
            BlocBuilder<FormDataCubit, FormDataState>(
              builder: (context, state) {
                if (state.data.validatorOf<Boolean>(formFieldPro).value) {
                  return Column(children: [
                    _SirenInput(),
                    const SizedBox(height: 8),
                    _IbanInput(),
                    const SizedBox(height: 8),
                  ]);
                }
                return const SizedBox.shrink();
              },
            ),
            const SizedBox(height: 8),
            _SignUpButton(),
            const SizedBox(height: 8),
            _DebugButton(),
          ],
        ),
      ),
    );
  }
}
